import inspect
import textwrap
import ast


def get_ast(obj):
    indented_program_txt = inspect.getsource(obj)
    program_txt = textwrap.dedent(indented_program_txt)
    return ast.parse(program_txt)



from hwtypes import BitVector

from .digital import Digital
from .bit import Bit
from .bits import Bits
from .bitutils import clog2, seq2int
from .circuit import coreir_port_mapping
from .conversions import array, as_bits, from_bits
from .generator import Generator2
from .interface import IO
from .protocol_type import MagmaProtocol
from .t import Type, In, Out
from .tuple import Product
from .wire import wire
from magma.mux import Mux


def mux(I, S, **kwargs):
    if isinstance(S, Type) and S.const():
        S = seq2int(S.bits())
    if isinstance(S, int):
        return I[S]
    T = type(I[0])
    return Mux(len(I), T, **kwargs)()(*I, S)


def slice(value: Bits, start: Bits, width: int):
    """
    Dynamic slice based off the value of `start` with constant `width`.

    Example:
    ```
    class TestSlice(m.Circuit):
        # Slice a 6 bit window of I using x as the start index
        IO = m.IO(
            I=m.In(m.Bits[10]),
            x=m.In(m.Bits[2]),
            O=m.Out(m.Bits[6])
        )

        io.O @= m.slice(io.I, start=io.x, width=6)
    ```

    Notes:
      * For now we only support slicing bits with bits index
      * We could add support for slicing Arrays
      * We could restrict index to be a UInt
    """
    # Construct an array where the index `i` is the slice of bits from `i` to
    # `i+width`, index into this array using `start`.
    return array([value[i:i + width] for i in range(len(value) - width)])[start]



from .passes import EditCircuitPass
from ..is_definition import isdefinition


def _terminate_unused(interface):
    terminated = False
    for port in interface.ports.values():
        if port.is_output() and not port.wired():
            terminated = True
            port.unused()
    return terminated


class TerminateUnusedPass(EditCircuitPass):
    def edit(self, circuit):
        if _terminate_unused(circuit.interface):
            circuit._is_definition = True
        if not isdefinition(circuit):
            return
        for inst in circuit.instances:
            _terminate_unused(inst.interface)



from .clock import WireClockPass
from .passes import DefinitionPass


class IRPass(DefinitionPass):
    def __init__(self, main):
        super().__init__(main)
        self.code = ""

        WireClockPass(main).run()

    def __call__(self, definition):
        self.code += repr(definition) + "\n\n"



from magma.array import Array
from magma.circuit import Circuit
from magma.digital import Digital
from magma.ref import DefnRef, InstRef, NamedRef, ArrayRef, TupleRef
from magma.t import Type
from magma.tuple import Tuple
from magma.view import InstView, PortView


def value_to_verilog_name(value):
    if isinstance(value, Array) and not issubclass(value.T, Digital):
        elems = ", ".join(value_to_verilog_name(t) for t in reversed(value))
        return f"'{{{elems}}}"
    elif isinstance(value, Tuple):
        raise NotImplementedError("Inlining unflattened tuple")
    elif isinstance(value, PortView):
        parent_name = value_to_verilog_name(value.parent)
        return f"{parent_name}.{verilog_name(value.port.name)}"
    elif isinstance(value, InstView):
        prefix = ""
        if isinstance(value.parent, Circuit):
            prefix = value.parent.name + "."
        else:
            prefix = value_to_verilog_name(value.parent) + "."
        return prefix + value.inst.name
    return verilog_name(value.name)


def verilog_name(name, inst_sep="."):
    if isinstance(name, DefnRef):
        return str(name)
    if isinstance(name, InstRef):
        return f"{name.inst.name}{inst_sep}{str(name)}"
    if isinstance(name, NamedRef):
        return str(name)
    if isinstance(name, ArrayRef):
        array_name = verilog_name(name.array.name)
        if issubclass(name.array.T, Digital):
            index = f"[{name.index}]"
        else:
            index = f"_{name.index}"
        return f"{array_name}{index}"
    if isinstance(name, TupleRef):
        tuple_name = verilog_name(name.tuple.name)
        index = name.index
        try:
            int(index)
            # python/coreir don't allow pure integer names
            index = f"_{index}"
        except ValueError:
            pass
        return f"{tuple_name}_{index}"
    raise NotImplementedError(name, type(name))


def convert_values_to_verilog_str(value):
    if isinstance(value, Type):
        # TODO: For now we assumed values aren't used elswhere, this
        # forces it to be connected to an instance, so temporaries will
        # appear in the code
        if value.is_inout() and not value.driven():
            raise NotImplementedError()
        if value.is_input() and not value.driven():
            # TODO: Could be driven after, but that will just override
            # this wiring so it's okay for now
            value.undriven()
        elif not value.is_input():
            # Always set to unused so the output wire isn't inlined
            value.unused()
        return value_to_verilog_name(value)
    if isinstance(value, PortView):
        return value_to_verilog_name(value)
    return str(value)



from functools import lru_cache
import weakref
from abc import ABCMeta
from .t import Kind, Direction, Type, In, Out
from .debug import debug_wire, get_callee_frame_info
from .compatibility import IntegerTypes
from .logging import root_logger
from magma.wire_container import Wire


_logger = root_logger()


class DigitalMeta(ABCMeta, Kind):
    _class_cache = weakref.WeakValueDictionary()

    def __new__(cls, name, bases, namespace, info=(None, None), **kwargs):
        # TODO: A lot of this code is shared with AbstractBitVectorMeta, we
        # should refactor to reuse
        if '_info_' in namespace:
            raise TypeError(
                'class attribute _info_ is reversed by the type machinery')

        direction = info[1]
        for base in bases:
            if getattr(base, 'is_directed', False):
                if direction is None:
                    direction = base.direction
                elif direction != base.direction:
                    raise TypeError(
                        "Can't inherit from multiple different directions")

        namespace['_info_'] = info[0], direction
        type_ = super().__new__(cls, name, bases, namespace, **kwargs)
        if direction is None:
            # class is unundirected so t.unundirected_t -> t
            type_._info_ = type_, direction
        elif info[0] is None:
            # class inherited from directed types so there is no unundirected_t
            type_._info_ = None, direction

        if not type_.is_directed:
            type_.VCC = type_[Direction.Out](name="VCC")
            type_.GND = type_[Direction.Out](name="GND")

        return type_

    def __getitem__(cls, direction: Direction) -> 'DigitalMeta':
        mcs = type(cls)
        try:
            return mcs._class_cache[cls, direction]
        except KeyError:
            pass

        if not isinstance(direction, Direction):
            raise TypeError('Direction of Digital must be an instance of '
                            'm.Direction')

        if cls.is_directed:
            if direction == direction.Undirected:
                return cls.undirected_t
            if direction == cls.direction:
                return cls
            else:
                return cls.undirected_t[direction]
        if direction == direction.Undirected:
            return cls

        bases = [cls]
        bases.extend(b[direction] for b in cls.__bases__ if isinstance(b, mcs))
        bases = tuple(bases)
        orig_name = cls.__name__
        class_name = '{}[{}]'.format(orig_name, direction.name)
        type_ = mcs(class_name, bases, {"orig_name": orig_name},
                    info=(cls, direction))
        type_.__module__ = cls.__module__
        mcs._class_cache[cls, direction] = type_
        return type_

    def __call__(cls, value=None, *args, **kwargs):
        if value is not None:
            if isinstance(value, (bool, IntegerTypes)):
                return cls.VCC if value else cls.GND
        result = super().__call__(*args, **kwargs)
        if value is not None:
            assert isinstance(value, Digital), type(value)
            result(value)
        return result

    def __repr__(cls):
        return str(cls)

    def __str__(cls):
        name = getattr(cls, "orig_name", cls.__name__)
        if cls.is_input():
            name = f"In({name})"
        elif cls.is_output():
            name = f"Out({name})"
        elif cls.is_inout():
            name = f"InOut({name})"
        return name

    @property
    def undirected_t(cls) -> 'DigitalMeta':
        t = cls._info_[0]
        if t is not None:
            return t
        else:
            raise AttributeError('type {} has no undirected_t'.format(cls))

    @property
    def direction(cls) -> int:
        return cls._info_[1]

    @property
    def is_directed(cls) -> bool:
        return cls.direction is not None

    def __len__(cls):
        return 1

    def qualify(cls, direction):
        return cls[direction]

    def __eq__(cls, rhs):
        return cls is rhs

    __hash__ = type.__hash__


class Digital(Type, metaclass=DigitalMeta):
    def __init__(self, value=None, name=None):
        super().__init__(name=name)
        self._wire = Wire(self)

    @classmethod
    def is_oriented(cls, direction):
        if not cls.is_directed:
            return direction == Direction.Undirected
        return cls.direction == direction

    @classmethod
    def is_clock(cls):
        return False

    def __call__(self, output):
        return self.wire(output, get_callee_frame_info())

    @debug_wire
    def wire(self, o, debug_info):
        i = self
        # promote integer types to LOW/HIGH
        if isinstance(o, IntegerTypes):
            o = HIGH if o else LOW

        if not isinstance(o, Digital):
            _logger.error(f'Cannot wire {i.debug_name} (type={type(i)}) to {o} '
                          f'(type={type(o)}) because {o.debug_name} is not a '
                          f'Digital', debug_info=debug_info)
            return

        i._wire.connect(o._wire, debug_info)
        i.debug_info = debug_info
        o.debug_info = debug_info

    def iswhole(self):
        return True

    def wired(self):
        return self._wire.wired()

    # return the input or output Bit connected to this Bit
    def trace(self):
        return self._wire.trace()

    # return the output Bit connected to this input Bit
    def value(self):
        return self._wire.value()

    def driven(self):
        return self._wire.driven()

    def driving(self):
        return self._wire.driving()

    @classmethod
    def unflatten(cls, value):
        if len(value) != 1 or not isinstance(value[0], Digital):
            raise TypeError("Can only convert from Bits[1] to Bit")
        return value[0]

    def flatten(self):
        return [self]

    def const(self):
        cls = type(self)
        return self is cls.VCC or self is cls.GND

    def unwire(i, o):
        i._wire.unwire(o._wire)

    @classmethod
    def flat_length(cls):
        return 1

    # Loam methods, TODO: rework this interface
    def getinst(self):
        t = self.trace()
        return t.name.inst if t is not None else None

    def getgpio(self):
        return self.getinst()

    @classmethod
    def is_mixed(cls):
        return False

    def __repr__(self):
        cls = type(self)
        if self is cls.VCC:
            return "VCC"
        if self is cls.GND:
            return "GND"
        return Type.__repr__(self)


VCC = Digital.VCC
GND = Digital.GND

HIGH = VCC
LOW = GND
